/***************************************************************************
 *  editbuffer.cc
 *
 *  Sun Nov 14 10:49:20 2004
 *  Copyright  2004 Ing. Stefano Salvi
 *  Email stefano@salvi.mn.it
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 

#include <string.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <regex.h>

#include <glib/gmacros.h>
#include <glib.h>
#include <gtkmm.h>
#include <libgnomemm.h>

#include "editbuffer.h"
#include "editwindow.h"
#include "messageinterface.h"

/*
 * B_G_List
 */

void B_G_List::append(undo_t* mydata)
{
	if (list == NULL) {
		tail = list = g_list_append(list, static_cast<undo_t*>(mydata));
	} else {
		tail = g_list_append(tail, static_cast<undo_t*>(mydata));
		tail = g_list_next(tail);
	}
}

undo_t* B_G_List::remove_link()
{
undo_t* el = tdata ();
GList *tmp = tail;
	tail = tail -> prev;
	list = g_list_remove_link(list, tmp);
	g_list_free_1 (tmp);
	return el;
}

/*
 * EditBuffer
 */

#ifdef GNP_DEBUG
/*
 * PRIVATE: undo_redo_list_print
 *
 * for debugging purposes only: prints the undo and redo lists.
 */
static void
undo_redo_list_print(doc_t *d, char *label)
{
	GList *lp;

	if (dbg_flags & DBG_UNDO) {
		printf("[%s] undolist: ", label);
		for (lp = d->undolist; lp; lp = lp->next) {
			printf("%s ", ((undo_t *)(lp->data))->text);
		}
		printf("\n");
		printf("[%s] redolist: ", label);
		for (lp = d->redolist; lp; lp = lp->next) {
			printf("%s ", ((undo_t *)(lp->data))->text);
		}
		printf("\n");
	}
} /* undo_redo_list_print */
#endif	/* GNP_DEBUG */

/* ******************************************************************************************* *
 * End Undo / redo code from GNP
 * ******************************************************************************************* */
/*
 * PUBLIC: undo_list_add
 *
 * adds an undo item to the undolist.  if there exists items on the redo list,
 * removes them first.
 */
void EditBuffer::undo_list_add(char *text, int start, int end, undo_t::undo_op_t op)
{
undo_t *undo;

	undo = new undo_t;
	undo->text = text;
	undo->start = start;
	undo->end = end;
	undo->changed = isChanged ();
	undo->op = op;
	
	/*
	 * if there's anything on the redo list, then we need to wipe out the
	 * redo list because we now have new undo items
	 */
	if (!redolist.empty ()) {
		undo_list_delete (redolist);
		parent -> redoSensitive (false);
	}

	/*
	 * add to the undo list.  note that since it's conceivable that the
	 * undo list could get extremely long, appending to the list on each
	 * key typed by the user would result in awful response time.  so what
	 * we do is we keep track of the tail (end) of the list, and always
	 * append using the tail.
	 *
	 * The reason why we even keep the tail around instead of just treating
	 * both undo and redo lists like a stack is because in the case of
	 * Undo/Redo "All", it's a lot faster to append one list to the other
	 * instead of popping off each item and adding them one at a time.
	 * This boneheaded method is what gedit uses (so don't do this).
	 */
	if (undolist.empty ())  {
		parent -> undoSensitive (true);
	}
	undolist.append(undo);

	undo_count_incr ();
} /* undo_list_add */

/*
 * PUBLIC: undo_list_delete
 *
 * used to remove either the undo list or the redo list.
 */
void EditBuffer::undo_list_delete(B_G_List &list)
{
	while (!list.empty ())
	{
		undo_list_head_chop(list);
	}
} /* undo_list_delete */

void EditBuffer::undo_count_decr()
{
	if (numundos > 0) {
		numundos--;
	}
} /* undo_count_decr */

void EditBuffer::undo_count_incr()
{
	numundos++;
#ifdef MAXUNDO
	if (prefs.maxundo > 0 && d->numundos > prefs.maxundo) {
		undo_list_head_chop(undolist);
		numundos--;
		g_assert(numundos == prefs.maxundo);
	}
#endif
} /* undo_count_incr */

/*
 * given an undo/redo list, remove the head element only.
 */
void EditBuffer::undo_list_head_chop(B_G_List &list)
{
undo_t *undo;
	undo = list.remove_link ();
	delete undo;
} /* list_head_chop */

/*
 * which: 0 = undo, 1 = redo
 */
void EditBuffer::undo_redo_common(undo_redo which)
{
undo_t *undo;
B_G_List *src;
B_G_List *dst;

#ifdef MAXUNDO
	if (prefs.maxundo == 0) {
		return;
	}
#endif
	/*
	 * we have two lists, the undo list, and the redo list.  here, figure
	 * out which one we're using, so that we know which list an item moves
	 * from, and which list it moves to.
	 */
	if (which == c_undo) {
		src = &undolist;
		dst = &redolist;
	} else {
		src = &redolist;
		dst = &undolist;
	}

	if (src -> empty ()) {
		return;
	}

//   undo_redo_list_print(d, "undo_redo_common before");
	disable_before_redo ();

	/* remove from src list, and place on dst list */
	undo = src -> remove_link ();
	/* 'set_point' raggiunge il punto voluto scrollando linea per linea, quindi
	 * controllo se la posizione di arrivio e' lontana e, se lo e', ci vado con
	 * 'realgoto' che ci va 'di botto'.
	 * Non uso 'get_point' per sapere la posizione corrente, perche' non rispecchia
	 * sempre la posizione del cursore (ma magari quella dell'ultimo 'set_point')
	 */
	if (which == c_undo) {	/* UNDO */
		switch (undo->op)
 		{
			case undo_t::UndoInsert :
				erase (get_iter_at_offset (undo->start),
					get_iter_at_offset (undo->end));
				placeCursor (undo->start);
				break;
			case undo_t::UndoReplace :
				undo_replace (undo);
				break;
			default:
				insert (get_iter_at_offset (undo->start), undo->text, undo->text + strlen (undo->text));
				placeCursor (undo->end);
				break;
		}
	} else {	/* REDO */
		switch (undo->op)
		{
			case undo_t::UndoInsert :
				insert (get_iter_at_offset (undo->start), undo->text, undo->text + strlen (undo->text));
				placeCursor (undo->end);
				break;
			case undo_t::UndoReplace :
				undo_replace(undo);
				break;
			default:
				erase (get_iter_at_offset (undo->start),
					get_iter_at_offset (undo->end));
				placeCursor (undo->start);
				break;
		}
	}
  
	if (dst -> empty ()) {	// Se prima era vuoto, attiva
		if (which == c_undo) {
			parent -> redoSensitive(true);
		} else {
			parent -> undoSensitive(true);
		}
	}
	dst -> append (undo);

	if (src -> empty ())  {
		if (which == c_undo) {
			parent -> undoSensitive(false);
		} else {
			parent -> redoSensitive(false);
		}
	}

	if (which == c_undo) {
		undo_count_decr ();	/* one less undo */
	} else {
		undo_count_incr ();		/* one more undo */
	}

	/* update the doc info label */
	if ((isChanged () != undo->changed) || which == c_redo)	/* redo is always changed */ {
		setChanged ((which == c_redo) ? true : undo->changed);
	}
	enable_after_redo ();
//  undo_redo_list_print(d, "undo_redo_common after");
} /* undo_redo_common */

void EditBuffer::setChanged (bool c)
{
	if (c != changedFlag) {
	char const *oldtitle = get_title ();
		if (c) {
		char *newtitle = new char [strlen (oldtitle) + 3];
			newtitle [0] = '*';
			newtitle [1] = ' ';
			strcpy (newtitle + 2,oldtitle);
			parent -> set_title (newtitle);
			delete newtitle;
		} else {
			parent -> set_title (oldtitle);
		}
	}
	changedFlag = c;
}

void EditBuffer::undo_replace (undo_t *undo)
{
Glib::ustring buf;
	buf = get_text (get_iter_at_offset (undo->start), get_iter_at_offset (undo->end));
	erase (get_iter_at_offset (undo->start), get_iter_at_offset (undo->end));
	insert (get_iter_at_offset (undo->start), undo->text, undo->text + strlen (undo->text));
	undo->end = undo->start + strlen (undo->text);
	placeCursor (undo->end);
	undo->text = EditWindow::newChars (buf);
} /* undo_replace */


#define GNPDBG_FILE(X)
void EditBuffer::placeCursor (int pos)
{
	Gtk::TextBuffer::iterator cursor = get_iter_at_offset (pos);
	place_cursor (cursor);
	parent -> setMessage ();
}

/*
 * PUBLIC: doc_delete_text_cb
 *
 * callback for text widget for text deletion from keyboard typing
 */
void EditBuffer::on_erase (const Gtk::TextBuffer::iterator& range_begin, const Gtk::TextBuffer::iterator& range_end)
{
Glib::ustring text;
	if (editing) {
		text = get_text (range_begin, range_end);
		undo_list_add (EditWindow::newChars (text), range_begin.get_offset(), range_end.get_offset(), undo_t::UndoDelete);
		parent -> setMessage ();
	}
	TextBuffer::on_erase(range_begin, range_end);
}

/*
 * PUBLIC: doc_insert_text_cb
 *
 * callback for text widget for text insertion from keyboard typing.  note that
 * this routine and the next one (doc_delete_text_cb()), gets invoked for EVERY
 * key typed by the user.
 */
void EditBuffer::on_insert (const Gtk::TextBuffer::iterator& pos, const Glib::ustring& text, int bytes)
{
	if (editing) {
	int poffs = pos.get_offset();
		undo_list_add(EditWindow::newChars (text), poffs, poffs + bytes, undo_t::UndoInsert);
		parent -> setMessage ();
	}
	TextBuffer::on_insert (pos, text, bytes);
} /* doc_insert_text_cb */

void EditBuffer::on_changed (void)
{
	if (!changedFlag) {
	char const *oldtitle = get_title ();
	char *newtitle = new char [strlen (oldtitle) + 3];
		newtitle [0] = '*';
		newtitle [1] = ' ';
		strcpy (newtitle + 2,oldtitle);
		parent -> set_title (newtitle);
		delete newtitle;
	}
	changedFlag = true;
	parent -> setMessage ();
	TextBuffer::on_changed ();
}

void EditBuffer::disable_before_redo()
{
  /*
   * need to stop the signal, because if we add/delete text, it will
   * generate the signal again!
   */
	editing = false;	// Disabilita raccolta modifiche
	begin_user_action ();
}


void EditBuffer::enable_after_redo()
{
	/* now re-establish the signals */
	editing = true;		// Riabilita raccolta modifiche
	end_user_action ();
	parent -> setMessage ();
}

int EditBuffer::file_save_execute(const char *file)
{
int fh;
int len;
char *bkname = new char [strlen (file) + 2];
struct stat stmp;

	if (file != filename) {	// Se il nome e' cambiato
		setFilename (new char [strlen (file) + 1]);
		strcpy (filename, file);
	}
	strcpy (bkname, filename);
	strcat (bkname,"~");

	if (!stat (filename,&stmp)) { // Il file esiste
		unlink (bkname);		// Elimina il vecchio backup
		rename (filename,bkname);	// Rinomina il file nel '~'
	}

	if ((fh = open(filename, O_WRONLY | O_CREAT | O_TRUNC, S_IREAD | S_IWRITE | S_IRGRP | S_IWGRP | 
		S_IROTH)) == EOF)  {
		MessageInterface::doYesDialog ("Salva ...", NULL, "Non posso registrare il file '%s'", filename);
		return false;
	}

  /*
   * this isn't very memory efficient; gtk_editable_get_chars()
   * (actually, it's gtk_text_get_chars()), returns the text in a newly
   * malloc'd buffer.  we're basically making another copy of the
   * document just to save it.  if the document is large, we might run
   * into problems.
   */
	Glib::ustring const buf = get_text(false);
	len = buf.length ();
	if (write (fh, buf.data () , len) != len) {
		perror("Error saving file");
		::close(fh);
		return false;
	}
	::close(fh);

	noChanged ();

	return true;
} /* file_save_execute */

/*
 * PRIVATE: file_open_execute (win_t *w, doc_t *d)
 *
 * actually do the low-level file opening and stuff it into the document's
 * text widget.
 */
int EditBuffer::file_open_execute (char *file)
{
size_t size;
off_t bytesread;
long num;
char *buf;
FILE *fp;
struct stat *sb;	

	setFilename (file);
	/* get file stats (e.g., file size is used by files list window) */
	sb = (struct stat *)g_malloc(sizeof(struct stat));
	if (stat(filename, sb) == -1) {
		g_free(sb);
		sb = NULL;
	}

	if (sb && sb->st_size > 0) {
	/* open file */
		if ((fp = fopen(filename, "r")) == NULL)  {
		char msg[255];
			sprintf(msg, "file_open_execute: could not open '%s'",
				filename);
	    	perror(msg);
			return 1;
		}

		size = sb->st_size;
		GNPDBG_FILE(("size %lu is ok (using half)\n", (gulong)size));
		if ((buf = (char *)malloc(size + 1)) == NULL) {
			perror("file_open_execute: unable to malloc read buf");
			fclose(fp);
			g_free(sb);
			return 1;
		}

		GNPDBG_FILE(("malloc'd %lu bytes for read buf\n",
			(gulong)(size+1)));

		/* buffer allocated, now actually read the file */
    	num = fread(buf, 1, size, fp);
    	if (num > 0) {
			buf[num]='\0';
			disable_before_redo();
			set_text ( buf, buf+num);
			enable_after_redo();
			bytesread += num;
	    } else if (ferror(fp)) {
			/* need to print warning/error message */
			perror("read error");
	    } else /* if (num == 0) */ {
				GNPDBG_FILE(("hmm... read zero bytes"));
    	}
		free(buf);
		// num = file_unlock(filename, fp);
		fclose(fp);
    } /* filesize > 0 */

    g_free(sb);
    /* misc settings */
	set_modified (false);
	changedFlag = false;
	placeCursor (0);
    parent -> set_title (get_title ());

    return 0;
}

char const *EditBuffer::get_title ()
{
char *base;
	if (!filename) {
		return "Senza titolo";
	}

	base = strrchr(filename, '/');
	if (base) {
    	return base + 1;
	} else {
    	return filename;
	}
}

int EditBuffer::strncstr (const std::string *haystack, const std::string *needle, 
  int caseSensitive)
{
int hpos = 0;
int hlen,nlen,tnlen, thlen;
const char *hs, *nd, *ths, *tnd;

  hs = haystack -> data ();
  nd = needle -> data ();
  hlen = haystack -> length ();
  nlen = needle -> length ();

  while (hlen)
  {
    for (ths = hs, tnd = nd, thlen = hlen, tnlen = nlen;
      ((caseSensitive) ? (*ths == *tnd) : (toupper (*ths) == toupper (*tnd))) &&
	tnlen && thlen ; ths ++, tnd ++, tnlen --, thlen --) ;
    if (!tnlen)
      return hpos;
    hs ++;
    hlen --;
    hpos ++;
  }

  return -1;
}

bool EditBuffer::search_execute(std::string const *find, std::string const *replace,
  searchFrom fromCursor, bool caseSensitive, bool regExp, int &endSearch)
{
int idx;
int startpos;		/* starting position/index */
int slen;				/* search text length */
int sblen;			/* search buffer length */
int rlen = 0;
char *rstr = NULL;
int cursorOffset, selMarkOffset;

	slen = find -> length ();
	if (slen < 1) {
		return false;	/* don't do anything if empty */
	}

	cursorOffset = get_insert () -> get_iter().get_offset();
	selMarkOffset = get_selection_bound () -> get_iter().get_offset();
	/* get search start position */
	if (fromCursor == fromStart) {
		startpos = 0;
	} else {
		startpos = MAX (cursorOffset,selMarkOffset);
	}

	/*
	 * get the text in the text widget...
	 * yikes, this will take a lot of memory.  perhaps in the future, GTK
	 * will have a better text widget.  it's either consume lots of memory,
	 * or have a search that's so slow and pathetic, it's not even useful.
	 */
	if (!endSearch) {
		if (fromCursor == selectOnly && selMarkOffset != cursorOffset) {
			startpos = MIN (cursorOffset, selMarkOffset);		// inizio -> inizio selezione
			endSearch = MAX (cursorOffset, selMarkOffset);	// Fine -> fine selezione
		} else {
			endSearch = end ().get_offset();	// Fine -> fine documento
		}
	}
	std::string const searchbuf = get_text (get_iter_at_offset (startpos), get_iter_at_offset (endSearch));
	sblen = searchbuf.length ();

	/* search for the text */
	if (regExp) {
#define SLENGTH 128
	regex_t rege;
	regmatch_t regm [10];	// Espressione e 10 sottoespressioni
	int err;
	char *f = EditWindow::newChars (*find);
		err = regcomp(&rege, f,caseSensitive ?
			REG_EXTENDED | REG_NEWLINE | REG_ICASE : REG_EXTENDED | REG_NEWLINE);
		if (err) {
		char *msg = new char [SLENGTH + 1];
			(void)regerror(err, &rege, msg, SLENGTH);
			MessageInterface::doYesDialog ("Espressione Regolare", NULL, msg);
			delete msg;
			return false;
		}
		delete f;
		char *s = EditWindow::newChars (searchbuf);

		err = regexec(&rege, s, 10, regm, 0);
		if (err) {
			if (err == REG_NOMATCH) {
				idx = -1; /* goes to 'no match found' */
			} else {
			char *msg = new char [SLENGTH + 1];
			(void)regerror(err, &rege, msg, SLENGTH);
				MessageInterface::doYesDialog ("Ricerca Fallita", NULL, msg);
				delete msg;
				delete s;
				regfree(&rege);
				return false;
			}
		} else {	// Testo trovato
			idx = regm[0].rm_so;
			slen = regm[0].rm_eo - regm[0].rm_so;
			if (replace) {	// Crea la stringa da sostituire, usando le sottoespressioni
			int len = replace -> length ();
			const char *x = replace -> data ();
			char *d;
				// Prima calcola la lunghezza
				rlen = 0;
				while (len --) {
					if (*x++ == '\\') {	// Sottostringa
						if (*x >= '0' && *x <= '9' && regm[*x - '0'].rm_so >= 0) {
							rlen += regm[*x - '0'].rm_eo - regm[*x - '0'].rm_so - 1;
							x ++;
						} else {
							rlen += 2;
							x ++;
							len --;
						}
					} else {
						rlen ++;
					}
				}
				// Poi crea la stringa con la sostituzione
				if (rlen) {
					len = replace -> length ();
					x = replace -> data ();
					d = rstr = new char [rlen + 1];
					while (len --) {
						if (*x == '\\') {	// Sottostringa
							x ++;	// Salta l' '\'
							if (*x >= '0' && *x <= '9' && regm[*x - '0'].rm_so >= 0) {
								strncpy (d, s + regm[*x - '0'].rm_so, regm[*x - '0'].rm_eo - regm[*x - '0'].rm_so);
								d += regm[*x - '0'].rm_eo - regm[*x - '0'].rm_so;
								x ++;
							} else {
								*d ++ = '\\';
								*d ++ = *x ++;
								len --;
							}
						} else {
							*d ++ = *x ++;
						}
						*d = 0;
					}
				}
			}
		}
		delete s;
		regfree(&rege);
#undef SLENGTH
	} else {	// Ricerca normale (non regex)
		idx = strncstr (&searchbuf, find, caseSensitive);
	}

	if (idx < 0) {
	char *msg = EditWindow::newChars (*find);
		MessageInterface::doYesDialog ("Testo non trovato", NULL, 
			" '%s' non trovato (raggiunta la fine del file) ", msg);
		delete msg;
		return false;
	}

	/* found the text, now go to the location and highlight it */
	idx += startpos;

	parent -> goTo (idx);	// si posiziona all'inizio e scrolla a quella posizione
	move_mark (get_selection_bound (), get_iter_at_offset (idx + slen));

	if (replace) {
		// Cancella la selezione
		erase (get_insert () -> get_iter(), get_selection_bound () -> get_iter());
		if (rlen) {	// Se ha testo da inserire
			if (regExp) {
				if (rstr) {	// Se esiste la stringa di sostituzione
					insert_at_cursor (rstr, rstr + strlen (rstr));
					delete rstr;  // Se e' una RE, allora la stringa e' allocata
				}
			} else {
				insert_at_cursor (* replace);
			}
		}
	}

	return true;
}
